#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <time.h>
#include <fitsio.h>
#include "pvo.hh"
#include "pvo-FITS.hh"
#include "pvo-utils.hh"
#include "pvo-block.hh"
#include "pvo-header.hh"

using namespace PVO::Block;
using namespace PVO::Header;

namespace PVO
{

  namespace FITS
  {
  
    PSRFits::PSRFits()
    {
      this->newFile = true;
      this->filestate = CLOSED;
      this->UT_initial = decToDMS(0.);
      this->verbose = true;
      this->writingData = true;
    }
    //-----------------------------------------------------------------------

    PSRFits::PSRFits(std::string name)
    {
      this->filestate = CLOSED;
      this->openFile(name);
      this->verbose = true;
      this->writingData = true;
    }
    //-----------------------------------------------------------------------

    void PSRFits::FITSError(std::string function, int status){
      /** 
       * An error message is printed, along with the relevant CFITSIO error
       * code and message.  
       * \param function The function from which the error was called.  
       * \param status The status value as returned by the CFITSIO function. 
       *  Generally, if status != 0, an error has occurred.
       */
      if(this->verbose){
	std::cerr << "\aFITSError <" << function 
		  << "> (status="    << status   <<"): ";
	fits_report_error(stderr, status);
      }
    }
    //------------------------------------------------------------
    template <typename T> int PSRFits::updateKeyword(std::string name, T value)
    {
      /**
       *   Updates the keyword "name" in the current header unit, giving
       *    it the value given by "value".
       *   A templated function that should work for all cases.
       */
      int status=0;
      fits_update_key(this->fptr, TINT, (char *)name.c_str(), &value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, int value){
      int status=0;
      fits_update_key(this->fptr, TINT, (char *)name.c_str(), 
		      &value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, long value){
      int status=0;
      fits_update_key(this->fptr, TLONG, (char *)name.c_str(), 
		      &value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, float value){
      int status=0;
      fits_update_key(this->fptr, TFLOAT, (char *)name.c_str(),
		      &value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, double value){
      int status=0;
      fits_update_key(this->fptr, TDOUBLE, (char *)name.c_str(), 
		      &value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, std::string value){
      int status=0;
      fits_update_key(this->fptr, TSTRING, (char *)name.c_str(), 
		      (char *)value.c_str(), NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    template <> int PSRFits::updateKeyword(std::string name, char *value){
      int status=0;
      fits_update_key(this->fptr, TSTRING, (char *)name.c_str(), 
		      value, NULL, &status);
      if(status) this->FITSError("PSRFits::updateKeyword",status);
      return status;
    }
    //------------------------------------------------------------------------
  
    template <typename T> int PSRFits::getKeyword(std::string name, T &value)
    {
      /**
       *   Finds the keyword "name" in the current header unit,
       *   returning the value in the "value" parameter.
       *   A templated function that should work for all cases.
       *  \param name The name of the keyword.
       *  \param value The returned value of the keyword.
       *  \return The status value from the fits_read_keyword function.
       */
      int status=0;
      char keywordValue[80], keywordComment[80];
      fits_read_keyword(this->fptr, (char *)name.c_str(), keywordValue, 
			keywordComment, &status);
      if(status) this->FITSError("PSRFits::getKeyword",status);
      else{
	std::stringstream ss;
	ss.str(keywordValue);
	ss >> value;
      }
      return status;
    }
    template int PSRFits::getKeyword<int>(std::string name, int &value);
    template int PSRFits::getKeyword<long>(std::string name, long &value);
    template int PSRFits::getKeyword<float>(std::string name, float &value);
    template int PSRFits::getKeyword<double>(std::string name, double &value);
    template <> int PSRFits::getKeyword<std::string>(std::string name, std::string &value){
      int status=0;
      char keywordValue[80], keywordComment[80];
      fits_read_keyword(this->fptr, (char *)name.c_str(), keywordValue, 
			keywordComment, &status);
      if(status) this->FITSError("PSRFits::getKeyword",status);
      else value = keywordValue;
      if(value[0]=='\'') value.erase(0,1);
      if(value[value.size()-1]=='\'') value.erase(value.size()-1,1);
      return status;
    }
    
  
    //------------------------------------------------------------------------

    void PSRFits::setCommandLine(char *executable, std::string inputFile)
    {
      /**
       *  Generates a string showing what the effective shell command
       *  was to generate the FITS file. The executable name has any
       *  leading directories stripped off. The input name inputFile
       *  is assigned the PSRFits::oldFilename string. No change is
       *  made to this name (i.e. no leading directories are removed).
       *
       * \param executable The name of the executable from the command
       *                    line. Can be argv[0] from 
       *                    main(int argc, char *argv[]).
       * \param inputFile The name of the input file.
       */


      //the stringstream to produce the commandLine string
      std::stringstream ss; 

      // strip off any leading directory from the executable.
      std::string exec = executable;
      unsigned int found = exec.rfind("/");
      if(found != std::string::npos){
	exec = exec.substr(found+1,exec.size()-1);
      }
      // write the executable to the string
      ss << exec << " ";

      // write the input filename after the executable.
      //  This avoids problems with wildcards or the use of the -D option.
      this->setOldFilename(inputFile);
      ss << this->oldFilename;

      // write the string to this->commandLine
      this->commandLine = ss.str();

    };
    //----------------------------------------------------------------------

    void PSRFits::setFITSname(DataBlock &block)
    {
      /**
       * Generates a string containing the name of the FITS file, based on the
       *  values of the initial date and time of the first observations in the 
       *  file, and (if the multibeam receiver was used) the beam number. 
       * The format is determined by the survey type. Possible formats are:
       * <ul> <li> S70_yymmdd_hhmmss.fits
       *      <li> PM_yymmdd_hhmmss_bb.fits </ul>
       * \param block The block that incorporates the survey type and
       *        allows us to work out what filename extension to use.
       * \return String containing the FITS filename.
       */
      std::string date;
      Header::SURVEY surveyType = block.getSurveyType();

      for(unsigned int i=0;i<this->DATE_initial.size();i++)
	if(DATE_initial[i]!='/') date += this->DATE_initial[i];

      std::string time = removeLeadingBlanks(this->UT_initial);
      time = decToDMS( dmsToDec(time), 0, "");
    
      std::stringstream output;
      if(surveyType == S70) output << "S70_";
      else if(surveyType == PM) output << "PM_";
      else if(surveyType == PH) output << "PH_";
      else PVOError("PSRFits::setFITSname","Unknown survey!\n");

      output << date << "_" << time;

      if((surveyType == PM)||(surveyType == PH))
	output << "_" 
	       << std::setw(2) << std::setfill('0') 
	       << this->beam_number;

      //      output << ".fits";
      // Get the extension.
      //  Base this on the MOVE parameter. 
      PVO::FITS::EXT_TYPE ext;
      std::string obsmode = block.getObs_mode();
      if(obsmode=="SEARCH") ext=SEARCH;
      else if(obsmode=="PSR") ext=TIMING;
      else{
	std::stringstream errmsg;
	errmsg << "Unknown OBS_MODE (" << obsmode << ")!\n";
	PVOError("PSRFits::setFITSname",errmsg.str());
	ext=SEARCH;
      }
      output << PVO::FITS::extension[ext];

      this->FITSname = output.str();
    
    }
    //-----------------------------------------------------------------------

    bool PSRFits::fileExists()
    {
      /**
       * Checks to see whether the file is exists. Uses the CFITSIO function
       * fits_file_exists, and the parameter PSRFits::FITSname.
       * \return True if the file exists, false otherwise.
       */
      int status = 0, exists;
      fits_file_exists(this->FITSname.c_str(), &exists, &status);
      if(status) this->FITSError("PSRFits::fileExists",status);
      if(exists>0) return true;
      else return false;
    }
    //--------------------------------------------------------------------------

    int PSRFits::openFile(std::string name)
    {
      /**
       *   If the file exists, it is opened.
       *   If not, it is created using the correct template file.
       *    In the latter case, the fix function is called, and the file is
       *    flagged as "new".
       *  \param name Filename of the fits file to be opened.
       */
      int status = 0, exists;
      fits_file_exists((char *)name.c_str(),&exists,&status);
      if(status) this->FITSError("PSRFits::openFile 1",status);
      status=0;
      if(exists<=0){
	char *template_file = (char *)getPSRFITSdefn().c_str();
	fits_create_template(&this->fptr, (char *)name.c_str(), template_file, &status);
	if(status) this->FITSError("PSRFits::openFile 2",status);
	else std::cerr << "Creating new file " << name << "\n";
	this->newFile = true;
      }
      else{
	fits_open_file(&this->fptr, name.c_str(), READWRITE, &status);
	if(status) this->FITSError("PSRFits::openFile 2",status);
	this->FITSname = name;
      }
      if(!status) {
	if(exists<=0) this->fixPSRFITS();
	this->gotoPHU();
	this->filestate = OPEN;
      }
      return status;
    }
    //--------------------------------------------------------------------------
  
    bool PSRFits::fitsFull(DataBlock &block)
    {
      /**
       * Checks to see if the FITS file has reached the point where it
       * needs to be closed. This should not happen until the later of:
       * <ul><li> MOVE flag is set to moving.
       *     <li> Number of samples is at least the minimum number
       *     determined by the survey.
       * </ul>
       * \return True if we have met *both* of these conditions. False
       * if we have not reached that point yet.
       */
      bool moveFlag = !block.telescopeOK();
      bool sampleFlag = (this->numSamples >= block.Survey().getMinSamples());
      return (moveFlag && sampleFlag);
    }
    //--------------------------------------------------------------------------
  
    int PSRFits::closeFile()
    {
      /**
       *   Closes the file if it is open. 
       *
       *   Sets the SCANLEN header based on the offs_sub column in the
       *   SUBINT table, so that it has the correct scanlen (not just
       *   the value that was requested).
       *
       *   Adjusts the filestate variable as well.
       */
      int status = 0;
      if(this->filestate == OPEN){

	float scanlen = this->calcScanLen();
	this->gotoPHU();
	this->updateKeyword("SCANLEN", scanlen );
	if(status) this->FITSError("PSRFits::closeFile",status);
	
	fits_close_file(this->fptr, &status);
	if(status) this->FITSError("PSRFits::closeFile",status);
	this->filestate = CLOSED;
      }
      return status;
    }

    //--------------------------------------------------------------------------
  
    int PSRFits::gotoPHU()
    {
      int status = 0;
      fits_movabs_hdu(this->fptr, 1, NULL, &status);    
      if(status) this->FITSError("PSRFits::gotoPHU",status);
      return status;
    }
    //--------------------------------------------------------------------------
  
    int PSRFits::gotoHistory()
    {
      int status = 0;
      char name[8];
      strcpy(name,"HISTORY");
      fits_movnam_hdu(this->fptr, BINARY_TBL, name, 0, &status);
      if(status) this->FITSError("PSRFits::gotoHistory",status);
      return status;
    }
    //--------------------------------------------------------------------------
  
    int PSRFits::deleteTable(char tableName[])
    {
      /** 
       * Deletes the named binary table -- first moving to the HDU then
       * deleting it. At completion, we have returned to the PHU.
       */ 
      int status = 0;
      int newHDUtype;
      fits_movnam_hdu(this->fptr, BINARY_TBL, tableName, 0, &status);
      if(status) this->FITSError("PSRFits::deleteTable",status);
      fits_delete_hdu(this->fptr, &newHDUtype, &status);
      if(status) this->FITSError("PSRFits::deleteTable",status);
      gotoPHU();
      return status;
    }
    //--------------------------------------------------------------------------
  
    int PSRFits::gotoSubint()
    {
      int status = 0;
      char name[8];
      strcpy(name,"SUBINT");
      fits_movnam_hdu(this->fptr, BINARY_TBL, name, 0, &status);
      if(status) this->FITSError("PSRFits::gotoSubint",status);
      return status;
    }
    //--------------------------------------------------------------------------
  
    int PSRFits::getColnum(char *name)
    {
      int status=0,colnum=0;
      fits_get_colnum(this->fptr, CASEINSEN, name, &colnum, &status);
      if(status) this->FITSError("getColnum", status);
      return colnum;
    }
  
    long PSRFits::getNumRows()
    {
      int status=0;
      long nrows=0;
      fits_get_num_rows(this->fptr, &nrows, &status);
      if(status) this->FITSError("getNumRows", status);
      return nrows;
    }
    //--------------------------------------------------------------------------
  
    template <class T> int PSRFits::writeColumnValue(char *name, T value)
    {
      /**
       *   Writes a new value in the column "name" in the current binary table.
       *   It adds a single value at the bottom of the table (ie. in the largest
       *    numbered row).
       *   A templated function that should work for all cases.
       */
      int status=0;
      int colnum = this->getColnum(name);
      long nrows = this->getNumRows();
      fits_write_col(this->fptr, TINT, colnum, nrows, 1, 1, &value, &status);
      if(status) this->FITSError("PSRFits::writeColumnValue", status);
      return status;
    }
    template int PSRFits::writeColumnValue<int>(char *name, int value);
    template <> int PSRFits::writeColumnValue(char *name, long value)
    {
      int status=0;
      int colnum = this->getColnum(name);
      long nrows = this->getNumRows();
      fits_write_col(this->fptr, TLONG, colnum, nrows, 1, 1, &value, &status);
      if(status) this->FITSError("PSRFits::writeColumnValue", status);
      return status;
    }
    template <> int PSRFits::writeColumnValue(char *name, float value)
    {
      int status=0;
      int colnum = this->getColnum(name);
      long nrows = this->getNumRows();
      fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 1, &value, &status);
      if(status) this->FITSError("PSRFits::writeColumnValue", status);
      return status;
    }
    template <> int PSRFits::writeColumnValue(char *name, double value)
    {
      int status=0;
      int colnum = this->getColnum(name);
      long nrows = this->getNumRows();
      fits_write_col(this->fptr, TDOUBLE, colnum, nrows, 1, 1, &value, &status);
      if(status) this->FITSError("PSRFits::writeColumnValue", status);
      return status;
    }
    template <> int PSRFits::writeColumnValue(char *name, char *value)
    {
      int status=0;
      int colnum = this->getColnum(name);
      long nrows = this->getNumRows();
      fits_write_col(this->fptr, TSTRING, colnum, nrows, 1, 1, &value, &status);
      if(status) this->FITSError("PSRFits::writeColumnValue", status);
      return status;
    }
     //--------------------------------------------------------------------------
  
    template <class T> int PSRFits::readColumnValue(char *name, int row, T *value)
    {
      /**
       *   Read the value in the column "name" in the current binary table.
       *   If the column is not present, the corresponding null value is returned. 
       *   A templated function that should work for all cases.
       */
      int status=0, anynul=0;
      int colnum = this->getColnum(name);
      T null=nullInt;
      fits_read_col(this->fptr, TINT, colnum, row, 1, 1, &null, value, &anynul, &status);
      if(status) this->FITSError("PSRFits::readColumnValue", status);
      return status;
    }
    template int PSRFits::readColumnValue<int>(char *name, int row, int *value);
    template <> int PSRFits::readColumnValue(char *name, int row, long *value)
    {
      int status=0, anynul=0;
      int colnum = this->getColnum(name);
      long null=nullInt;
      fits_read_col(this->fptr, TLONG, colnum, row, 1, 1, &null, value, &anynul, &status);
      if(status) this->FITSError("PSRFits::readColumnValue", status);
      return status;
    }
    template <> int PSRFits::readColumnValue(char *name, int row, float *value)
    {
      int status=0, anynul=0;
      int colnum = this->getColnum(name);
      float null=nullFloat;
      fits_read_col(this->fptr, TFLOAT, colnum, row, 1, 1, &null, value, &anynul, &status);
      if(status) this->FITSError("PSRFits::readColumnValue", status);
      return status;
    }
    template <> int PSRFits::readColumnValue(char *name, int row, double *value)
    {
      int status=0, anynul=0;
      int colnum = this->getColnum(name);
      double null=nullFloat;
      fits_read_col(this->fptr, TDOUBLE, colnum, row, 1, 1, &null, value, &anynul, &status);
      if(status) this->FITSError("PSRFits::readColumnValue", status);
      return status;
    }
    template <> int PSRFits::readColumnValue(char *name, int row, char *value)
    {
      int status=0, anynul=0;
      int colnum = this->getColnum(name);
      char *null = new char[nullString.size()+1];
      strcpy(null,nullString.c_str());
      fits_read_col(this->fptr, TSTRING, colnum, row, 1, 1, null, value, &anynul, &status);
      if(status) this->FITSError("PSRFits::readColumnValue", status);
      return status;
    }
    //--------------------------------------------------------------------------

    int PSRFits::startNewRow()
    {
      /**
       *    Adds a new row to the bottom of the current binary table.
       *    Also adds the row number as the ISUBINT column value.
       */
      int nrows = this->getNumRows();
      int status = 0;
      fits_insert_rows(this->fptr, nrows, 1, &status);
      if(status) this->FITSError("PSRFits::startNewRow", status);
      return status;
    }
    //--------------------------------------------------------------------------

    void PSRFits::startNewSubintRow()
    {
      /**
       *    Adds a new row to the bottom of the current binary table.
       *    Also adds the row number as the ISUBINT column value.
       */
      this->startNewRow();
      this->writeColumnValue("ISUBINT", this->getNumRows());
    }
    //--------------------------------------------------------------------------

    float PSRFits::calcScanLen()
    {
      /**
       *  Calculate the actual scan length of the FITS file by looking
       *  at the offs_sub column in the SUBINT table. We look at the
       *  last value, and add half the BLK_SEC value to it to get the
       *  total elapsed time.
       *
       * \return The total scan length of the FITS file.
       */

      this->gotoSubint();
      int numrows = this->getNumRows();
      float scanlen,oldscanlen,blksec;
      int status = this->readColumnValue<float>("OFFS_SUB",numrows,&scanlen);
      if(status) this->FITSError("PSRFits::calcScanLen", status);
      status = this->readColumnValue<float>("OFFS_SUB",numrows-1,&oldscanlen);
      if(status) this->FITSError("PSRFits::calcScanLen", status);
      blksec = scanlen - oldscanlen;
      scanlen += 0.5 * blksec;
      this->gotoPHU();
      return scanlen;

    }


  }

}
