#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <unistd.h>
#include <getopt.h>
#include <glob.h>
#include "pvo.hh"
#include "pvo-block.hh"
#include "pvo-header.hh"
#include "pvo-FITS.hh"
#include "pvo-utils.hh"

using namespace PVO;
using namespace PVO::Header;
using namespace PVO::Block;
using namespace PVO::FITS;
using std::string;

int main(int argc, char * argv[])
{
  /** 
   * PVOConvert is a function that converts pulsar survey data into a
   * standard PSRFITS format.
   *
   * This program is a command-line program that accepts an existing
   * file as an argument. It works out what survey the data is from
   * (based initially on the first few bytes of the file, and also
   * using the input flags -- see below), and uses this information to
   * interpret the header blocks.
   *
   * The binary data is written to the SUBINT binary table extension,
   * and a record of the command-line statement is written to the
   * HISTORY table. No other tables are written.
   *
   * The format of usage is: <br> 
   * <tt>> pvoconvert FILE </tt><br>
   * Wildcards or multiple files are accepted. To run on all files in a
   * directory, use the <tt>-D</tt> option:<br>
   * <tt>> pvoconvert -D DIR </tt><br>
   * If the input FILE is from the Parkes High Latitude Multibeam
   * Survey (PH, Burgay et al. 2006), one needs to use the <tt>-H</tt>
   * flag to use the PH settings (for integration time etc), rather
   * than those of the Multibeam survey (PM, Manchester et
   * al. 2001):<br>
   * <tt>> pvoconvert -H FILE </tt><br>
   * Using the <tt>-v</tt> option means all error messages will be
   * displayed. Entering<br>
   * <tt>> pvosummary -h</tt><br>
   * will give a help message, using PVO::pvoconvert_usage_text.
   *
   */

  std::vector<std::string> namelist;
  bool verbFlag = false;
  bool usingHighLatitude = false;
  bool writingData = true;
  unsigned int i;

  if(argc==1){
    std::cout << PVO::pvoconvert_usage_text_short;
    return 1;
  }

  char c;
  while (true) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, 0, 'h'},
      {"highlat", 0, 0, 'H'},
      {"directory", 1, 0, 'D'},
      {"test", 1, 0, 't'},
      {"verbose", 1, 0, 'V'},
      {"version", 1, 0, 'v'},
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, "D:HVvht",
		     long_options, &option_index);

    if (c == -1) break;

    std::string directory,filenames;
    glob_t globbuf1,globbuf2;

    switch (c) {               
    case 0:
      // This is where we deal with -- options that don't have a
      // single letter equivalent
      break;
    case 'D': // use all files in the specified directory.
      directory = optarg;
      globbuf1.gl_offs = 0;
      globbuf2.gl_offs = 0;
      filenames = directory + "/*";
      glob(filenames.c_str(), GLOB_DOOFFS, NULL, &globbuf1);
      namelist.resize(globbuf1.gl_pathc);
      for(i=0;i<globbuf1.gl_pathc+globbuf1.gl_offs;i++) {
	namelist[i] = globbuf1.gl_pathv[i];
      }
      break;
    case 'v': // return version number
      std::cout << PVO::PVOCONVERT << " version " << PVO::PVO_VERSION << "\n";
      return 0;
      break;
    case 'V': // use verbose mode
      verbFlag = true;
      if(argc==2){
	std::cout << PVO::pvoconvert_usage_text_short;
	return 1;
      }
      break;
    case 't': // test mode -- don't write any data, just report the filenames.
      writingData = false;
      break;
    case 'h': // give help statement
      std::cout << PVO::pvoconvert_usage_text;
      return 0;
      break;
    case 'H': // High Latitude survey
      usingHighLatitude = true;
      break;
    case '?': // invalid option
    case ':': // missing an argument
      std::cout <<  PVO::pvoconvert_usage_text_short;
      return 1;
      break;
    default:
      std::cerr << "ERROR: getopt returned character code " <<  c << "\n";
      break;
    }
  }

  if (optind < argc) { 
    // There is a non-option argument -- must be the file to access
    while (optind < argc) namelist.push_back( argv[optind++] );
  }

  if(!writingData){
    std::cout << "List of files that will be produced if -t is not used: \n";
    if(namelist.size()>1)
      std::cout << "  (grouped according to input file)\n";
  }

  for(i=0; i<namelist.size(); i++){

    std::string fname = namelist[i];
    bool doThisOne = true;

    SURVEY currentSurvey = idSurvey(fname);
    if(currentSurvey == NONEXISTENT) {
      std::cout << "ERROR: File " << fname <<" does not exist. Skipping this.\n";
      doThisOne = false;
    }

    if(usingHighLatitude){
      if(currentSurvey == PM) currentSurvey = PH;
      else if(currentSurvey!=DUD){
	std::cout << "WARNING: " << fname << " is a " << surveyNames[currentSurvey] <<" file.\n"
		  << " High Latitude survey settings are NOT being used.\n";
// 	doThisOne = false;
      }
    }

    std::string file_loc = "";
    int cwdSize=200;
    char *cwd = new char[200];
    if(fname[0]!='/'){
//       file_loc = get_current_dir_name();
      getcwd(cwd,cwdSize);
      file_loc=cwd;
      file_loc += "/";
    }
    file_loc += fname;
    if(doThisOne) {
      std::cout << "Input file " << file_loc <<" : ";
      if(currentSurvey == DUD){
	std::cout << "  Skipping this file.\n";
	doThisOne = false;
      }
      else {
	std::cout << surveyNames[currentSurvey]<<"\n";
      }
    }

    if(doThisOne){

      std::ifstream fin;
      fin.open(fname.c_str(), std::ios::binary);

      bool eof;
      DataBlock block(currentSurvey);
      block.readBlock(fin);
  
      int newBeam = block.getKeyword<int>(IBEAM);
      int oldBeam;
      bool newMoveFlag= block.telescopeOK();
      bool oldMoveFlag;

      PSRFits fitsfile;
      fitsfile.setWritingData(writingData);
      fitsfile.setVerbosity(verbFlag);
      //      fitsfile.setOldFilename(fname);
      fitsfile.setCommandLine(argv[0],fname);

      if(fitsfile.initialise(block) == FAILURE){
	std::cerr << "Could not open PSRFITS file "
		  << fitsfile.getFITSname()
		  << ".\nExiting...\n";
	return 1;
      }

      do{

	if(fitsfile.getFilestate()==OPEN) fitsfile.writeSubintRow(block);

	oldBeam = newBeam;
	oldMoveFlag = newMoveFlag;
 	
	// We've done all we can for this block. Read a new one.
	eof = !block.readBlock(fin);

	if(!eof){ // if the file is still open
      
	  bool fileFull = fitsfile.fitsFull(block) ;

	  // get info on the new block
	  newBeam     = block.getKeyword<int>(IBEAM);
	  newMoveFlag = block.telescopeOK();
      
	  if((newBeam != oldBeam) || fileFull ){
	    //if we've moved, or it's a new beam, close the file.
	    fitsfile.closeFile();
	  }

	  if(newMoveFlag){

	    if(!oldMoveFlag){
	      // if this is true, then we were moving but now are not.
	      // this means it is a new file.
	      fitsfile.setNewFile();
	    }

	    if( fitsfile.getFilestate()==CLOSED ){
	      // if there is no file open, but we are not moving, 
	      //  open a new file
	      fitsfile.initialise(block);
	  
	    }

	  }

	}
	
      }while(!eof);
  
      fitsfile.closeFile();
      fin.close();

    } // end of if(doThisOne) clause

  } // end loop over namelist

  return 0;

}
